package handler

import (
	"fmt"
	"github.com/bytedream/sshoneypot/sshoneypot/info"
	"github.com/pborman/getopt"
	"math/rand"
	"mime"
	"net"
	"regexp"
	"strconv"
	"strings"
	"time"
)

func randomIp() string {
	return fmt.Sprintf("%d.%d.%d.%d", rand.Intn(254), rand.Intn(254), rand.Intn(254), rand.Intn(254))
}

func Ifconfig(cmdInfo *info.Info) {
	for _, arg := range cmdInfo.Args {
		if arg == "--help" {
			cmdInfo.Writeln(`Usage:
  ifconfig [-a] [-v] [-s] <interface> [[<AF>] <address>]
  [add <address>[/<prefixlen>]]
  [del <address>[/<prefixlen>]]
  [[-]broadcast [<address>]]  [[-]pointopoint [<address>]]
  [netmask <address>]  [dstaddr <address>]  [tunnel <address>]
  [outfill <NN>] [keepalive <NN>]
  [hw <HW> <address>]  [mtu <NN>]
  [[-]trailers]  [[-]arp]  [[-]allmulti]
  [multicast]  [[-]promisc]
  [mem_start <NN>]  [io_addr <NN>]  [irq <NN>]  [media <type>]
  [txqueuelen <NN>]
  [[-]dynamic]
  [up|down] ...

  <HW>=Hardware Type.
  List of possible hardware types:
    loop (Local Loopback) slip (Serial Line IP) cslip (VJ Serial Line IP) 
    slip6 (6-bit Serial Line IP) cslip6 (VJ 6-bit Serial Line IP) adaptive (Adaptive Serial Line IP) 
    ash (Ash) ether (Ethernet) ax25 (AMPR AX.25) 
    netrom (AMPR NET/ROM) rose (AMPR ROSE) tunnel (IPIP Tunnel) 
    ppp (Point-to-Point Protocol) hdlc ((Cisco)-HDLC) lapb (LAPB) 
    arcnet (ARCnet) dlci (Frame Relay DLCI) frad (Frame Relay Access Device) 
    sit (IPv6-in-IPv4) fddi (Fiber Distributed Data Interface) hippi (HIPPI) 
    irda (IrLAP) ec (Econet) x25 (generic X.25) 
    eui64 (Generic EUI-64) 
  <AF>=Address family. Default: inet
  List of possible address families:
    unix (UNIX Domain) inet (DARPA Internet) inet6 (IPv6) 
    ax25 (AMPR AX.25) netrom (AMPR NET/ROM) rose (AMPR ROSE) 
    ipx (Novell IPX) ddp (Appletalk DDP) ec (Econet) 
    ash (Ash) x25 (CCITT X.25)`)
			return
		}
	}
	cmdInfo.Writeln(`eth0: flags=4163<UP,BROADCAST,RUNNING,MULTICAST>  mtu 1500
        inet 172.17.0.2  netmask 255.255.0.0  broadcast 172.17.255.255
        ether 02:42:ac:11:00:02  txqueuelen 0  (Ethernet)
        RX packets 13  bytes 1182 (1.1 KiB)
        RX errors 0  dropped 0  overruns 0  frame 0
        TX packets 0  bytes 0 (0.0 B)
        TX errors 0  dropped 0 overruns 0  carrier 0  collisions 0

		lo: flags=73<UP,LOOPBACK,RUNNING>  mtu 65536
				inet 127.0.0.1  netmask 255.0.0.0
				loop  txqueuelen 1000  (Local Loopback)
				RX packets 0  bytes 0 (0.0 B)
				RX errors 0  dropped 0  overruns 0  frame 0
				TX packets 0  bytes 0 (0.0 B)
				TX errors 0  dropped 0 overruns 0  carrier 0  collisions 0`)
}

type Ping struct {
	info.SimpleHandler
}

func (ping Ping) Handle(cmdInfo *info.Info) {
	opts := getopt.New()

	count := opts.Int('c', 4, "count")
	interval := opts.String('i', "1", "interval")
	packetsize := opts.Int('s', 56, "packetsize")
	opts.Getopt(cmdInfo.FullArgs, nil)

	duration, err := strconv.ParseFloat(*interval, 64)
	if err != nil {
		cmdInfo.Writelnf("%s: bad timing interval", cmdInfo.Command)
		return
	}

	address := opts.Arg(opts.NArgs() - 1)
	ip := address
	if net.ParseIP(address) == nil {
		ip = randomIp()
	}

	cmdInfo.Writelnf("PING %s (%s) %d data bytes.", address, ip, *packetsize)
	for i := 0; i < *count; i++ {
		resultTime := rand.Float64() * 10
		cmdInfo.Writelnf("%d bytes from %s (%s): icmp_seq=%d ttl=119 time=%.2f", *packetsize + 8, address, ip, i + 1, resultTime)
		time.Sleep(time.Duration(duration) * time.Second)
	}
}

func (ping Ping) help() string {
	return `Usage: ping [-aAbBdDfhLnOqrRUvV64] [-c count] [-i interval] [-I interface]
            [-m mark] [-M pmtudisc_option] [-l preload] [-p pattern] [-Q tos]
            [-s packetsize] [-S sndbuf] [-t ttl] [-T timestamp_option]
            [-w deadline] [-W timeout] [hop1 ...] destination
	Usage: ping -6 [-aAbBdDfhLnOqrRUvV] [-c count] [-i interval] [-I interface]
				 [-l preload] [-m mark] [-M pmtudisc_option]
				 [-N nodeinfo_option] [-p pattern] [-Q tclass] [-s packetsize]
				 [-S sndbuf] [-t ttl] [-T timestamp_option] [-w deadline]
				 [-W timeout] destination`
}

type Wget struct {
	info.SimpleHandler
}

func (wget Wget) Handle(cmdInfo *info.Info) {
	if _, urls, ok := basicFileCommandParser(cmdInfo, wget.help()); ok {
		regex := regexp.MustCompile("/((([A-Za-z]{3,9}:(?:\\/\\/)?)(?:[-;:&=\\+\\$,\\w]+@)?[A-Za-z0-9.-]+|(?:www.|[-;:&=\\+\\$,\\w]+@)[A-Za-z0-9.-]+)((?:\\/[\\+~%\\/.\\w-_]*)?\\??(?:[-\\+=&;%@.\\w_]*)#?(?:[\\w]*))?)/")
		for _, url := range urls {
			var port uint32
			schemeUrl := url

			if strings.HasPrefix(url, "http://") {
				port = 80
			} else if strings.HasPrefix(url, "https://") {
				port = 433
			} else {
				schemeUrl = "http://" + url
				port = 433
			}

			currentTime := time.Now().Format("2006-01-02 15:04:05")

			if regex.MatchString(url) {
				// generates a random ip

				mimeType := mime.TypeByExtension(url)
				if mimeType == "" {
					mimeType = "text/html"
				}

				ip := randomIp()

				cmdInfo.Writelnf(`--%s--  %s
					Connecting to %s (%s)|%s|:%d... connected."
					HTTP request sent, awaiting response... 200 OK
					Length: unspecified [%s]
					Saving to: '%s'`, currentTime, schemeUrl, url, url, ip, port, mimeType, url)

				// TODO: download bar

				// sleeps a random time to imitate the fetching process
				time.Sleep(time.Duration(rand.Intn(5000)) * time.Millisecond)

				if cmdInfo.Terminal.Width >= 60 {
					loadingBar := 1
					if cmdInfo.Terminal.Width > 60 {
						loadingBar += int((cmdInfo.Terminal.Width - 60) / 2)
					}
				}
			} else {
				cmdInfo.Writelnf(`--%s--  %s
					Resolving %s (%s)... failed: Name or service not known.
					wget: unable to resolve host address '%s'`, currentTime, schemeUrl, url, url, url)
			}
		}
	}
}

func (wget Wget) help() string {
	return `GNU Wget 1.20.1, a non-interactive network retriever.
	Usage: wget [OPTION]... [URL]...
	
	Mandatory arguments to long options are mandatory for short options too.
	
	Startup:
	  -V,  --version                   display the version of Wget and exit
	  -h,  --help                      print this help
	  -b,  --background                go to background after startup
	  -e,  --execute=COMMAND           execute a '.wgetrc'-style command
	
		Logging and input file:
		-o,  --output-file=FILE          log messages to FILE
		-a,  --append-output=FILE        append messages to FILE
		-d,  --debug                     print lots of debugging information
		-q,  --quiet                     quiet (no output)
		-v,  --verbose                   be verbose (this is the default)
		-nv, --no-verbose                turn off verboseness, without being quiet
		--report-speed=TYPE         output bandwidth as TYPE.  TYPE can be bits
		-i,  --input-file=FILE           download URLs found in local or external FILE
		-F,  --force-html                treat input file as HTML
		-B,  --base=URL                  resolves HTML input-file links (-i -F)
		relative to URL
		--config=FILE               specify config file to use
		--no-config                 do not read any config file
		--rejected-log=FILE         log reasons for URL rejection to FILE
	
		Download:
		-t,  --tries=NUMBER              set number of retries to NUMBER (0 unlimits)
		--retry-connrefused         retry even if connection is refused
		--retry-on-http-error=ERRORS    comma-separated list of HTTP errors to retry
		-O,  --output-document=FILE      write documents to FILE
		-nc, --no-clobber                skip downloads that would download to
		existing files (overwriting them)
		--no-netrc                  don't try to obtain credentials from .netrc
		-c,  --continue                  resume getting a partially-downloaded file
		--start-pos=OFFSET          start downloading from zero-based position OFFSET
		--progress=TYPE             select progress gauge type
		--show-progress             display the progress bar in any verbosity mode
		-N,  --timestamping              don't re-retrieve files unless newer than
		local
		--no-if-modified-since      don't use conditional if-modified-since get
		requests in timestamping mode
		--no-use-server-timestamps  don't set the local file's timestamp by
		the one on the server
		-S,  --server-response           print server response
		--spider                    don't download anything
		-T,  --timeout=SECONDS           set all timeout values to SECONDS
		--dns-timeout=SECS          set the DNS lookup timeout to SECS
		--connect-timeout=SECS      set the connect timeout to SECS
		--read-timeout=SECS         set the read timeout to SECS
		-w,  --wait=SECONDS              wait SECONDS between retrievals
		--waitretry=SECONDS         wait 1..SECONDS between retries of a retrieval
		--random-wait               wait from 0.5*WAIT...1.5*WAIT secs between retrievals
		--no-proxy                  explicitly turn off proxy
		-Q,  --quota=NUMBER              set retrieval quota to NUMBER
		--bind-address=ADDRESS      bind to ADDRESS (hostname or IP) on local host
		--limit-rate=RATE           limit download rate to RATE
		--no-dns-cache              disable caching DNS lookups
		--restrict-file-names=OS    restrict chars in file names to ones OS allows
		--ignore-case               ignore case when matching files/directories
		-4,  --inet4-only                connect only to IPv4 addresses
		-6,  --inet6-only                connect only to IPv6 addresses
		--prefer-family=FAMILY      connect first to addresses of specified family,
		one of IPv6, IPv4, or none
		--user=USER                 set both ftp and http user to USER
		--password=PASS             set both ftp and http password to PASS
		--ask-password              prompt for passwords
		--use-askpass=COMMAND       specify credential handler for requesting
		username and password.  If no COMMAND is
		specified the WGET_ASKPASS or the SSH_ASKPASS
		environment variable is used.
		--no-iri                    turn off IRI support
		--local-encoding=ENC        use ENC as the local encoding for IRIs
		--remote-encoding=ENC       use ENC as the default remote encoding
		--unlink                    remove file before clobber
		--xattr                     turn on storage of metadata in extended file attributes
	
		Directories:
		-nd, --no-directories            don't create directories
		-x,  --force-directories         force creation of directories
		-nH, --no-host-directories       don't create host directories
		--protocol-directories      use protocol name in directories
		-P,  --directory-prefix=PREFIX   save files to PREFIX/..
		--cut-dirs=NUMBER           ignore NUMBER remote directory components
	
		HTTP options:
		--http-user=USER            set http user to USER
		--http-password=PASS        set http password to PASS
		--no-cache                  disallow server-cached data
		--default-page=NAME         change the default page name (normally
		this is 'index.html'.)
		-E,  --adjust-extension          save HTML/CSS documents with proper extensions
		--ignore-length             ignore 'Content-Length' header field
		--header=STRING             insert STRING among the headers
		--compression=TYPE          choose compression, one of auto, gzip and none. (default: none)
		--max-redirect              maximum redirections allowed per page
		--proxy-user=USER           set USER as proxy username
		--proxy-password=PASS       set PASS as proxy password
		--referer=URL               include 'Referer: URL' header in HTTP request
		--save-headers              save the HTTP headers to file
		-U,  --user-agent=AGENT          identify as AGENT instead of Wget/VERSION
		--no-http-keep-alive        disable HTTP keep-alive (persistent connections)
		--no-cookies                don't use cookies
		--load-cookies=FILE         load cookies from FILE before session
		--save-cookies=FILE         save cookies to FILE after session
		--keep-session-cookies      load and save session (non-permanent) cookies
		--post-data=STRING          use the POST method; send STRING as the data
		--post-file=FILE            use the POST method; send contents of FILE
		--method=HTTPMethod         use method "HTTPMethod" in the request
		--body-data=STRING          send STRING as data. --method MUST be set
		--body-file=FILE            send contents of FILE. --method MUST be set
		--content-disposition       honor the Content-Disposition header when
		choosing local file names (EXPERIMENTAL)
		--content-on-error          output the received content on server errors
		--auth-no-challenge         send Basic HTTP authentication information
		without first waiting for the server's
		challenge
	
		HTTPS (SSL/TLS) options:
		--secure-protocol=PR        choose secure protocol, one of auto, SSLv2,
		SSLv3, TLSv1, TLSv1_1, TLSv1_2 and PFS
		--https-only                only follow secure HTTPS links
		--no-check-certificate      don't validate the server's certificate
		--certificate=FILE          client certificate file
		--certificate-type=TYPE     client certificate type, PEM or DER
		--private-key=FILE          private key file
		--private-key-type=TYPE     private key type, PEM or DER
		--ca-certificate=FILE       file with the bundle of CAs
		--ca-directory=DIR          directory where hash list of CAs is stored
		--crl-file=FILE             file with bundle of CRLs
		--pinnedpubkey=FILE/HASHES  Public key (PEM/DER) file, or any number
		of base64 encoded sha256 hashes preceded by
		'sha256//' and separated by ';', to verify
		peer against
	
		--ciphers=STR           Set the priority string (GnuTLS) or cipher list string (OpenSSL) directly.
		Use with care. This option overrides --secure-protocol.
		The format and syntax of this string depend on the specific SSL/TLS engine.
		HSTS options:
		--no-hsts                   disable HSTS
		--hsts-file                 path of HSTS database (will override default)
	
		FTP options:
		--ftp-user=USER             set ftp user to USER
		--ftp-password=PASS         set ftp password to PASS
		--no-remove-listing         don't remove '.listing' files
		--no-glob                   turn off FTP file name globbing
		--no-passive-ftp            disable the "passive" transfer mode
		--preserve-permissions      preserve remote file permissions
		--retr-symlinks             when recursing, get linked-to files (not dir)
	
		FTPS options:
		--ftps-implicit                 use implicit FTPS (default port is 990)
		--ftps-resume-ssl               resume the SSL/TLS session started in the control connection when
		opening a data connection
		--ftps-clear-data-connection    cipher the control channel only; all the data will be in plaintext
		--ftps-fallback-to-ftp          fall back to FTP if FTPS is not supported in the target server
		WARC options:
		--warc-file=FILENAME        save request/response data to a .warc.gz file
		--warc-header=STRING        insert STRING into the warcinfo record
		--warc-max-size=NUMBER      set maximum size of WARC files to NUMBER
		--warc-cdx                  write CDX index files
		--warc-dedup=FILENAME       do not store records listed in this CDX file
		--no-warc-compression       do not compress WARC files with GZIP
		--no-warc-digests           do not calculate SHA1 digests
		--no-warc-keep-log          do not store the log file in a WARC record
		--warc-tempdir=DIRECTORY    location for temporary files created by the
		WARC writer
	
		Recursive download:
		-r,  --recursive                 specify recursive download
		-l,  --level=NUMBER              maximum recursion depth (inf or 0 for infinite)
		--delete-after              delete files locally after downloading them
		-k,  --convert-links             make links in downloaded HTML or CSS point to
		local files
		--convert-file-only         convert the file part of the URLs only (usually known as the basename)
		--backups=N                 before writing file X, rotate up to N backup files
		-K,  --backup-converted          before converting file X, back up as X.orig
		-m,  --mirror                    shortcut for -N -r -l inf --no-remove-listing
		-p,  --page-requisites           get all images, etc. needed to display HTML page
		--strict-comments           turn on strict (SGML) handling of HTML comments
	
		Recursive accept/reject:
		-A,  --accept=LIST               comma-separated list of accepted extensions
		-R,  --reject=LIST               comma-separated list of rejected extensions
		--accept-regex=REGEX        regex matching accepted URLs
		--reject-regex=REGEX        regex matching rejected URLs
		--regex-type=TYPE           regex type (posix|pcre)
		-D,  --domains=LIST              comma-separated list of accepted domains
		--exclude-domains=LIST      comma-separated list of rejected domains
		--follow-ftp                follow FTP links from HTML documents
		--follow-tags=LIST          comma-separated list of followed HTML tags
		--ignore-tags=LIST          comma-separated list of ignored HTML tags
		-H,  --span-hosts                go to foreign hosts when recursive
		-L,  --relative                  follow relative links only
		-I,  --include-directories=LIST  list of allowed directories
		--trust-server-names        use the name specified by the redirection
		URL's last component
		-X,  --exclude-directories=LIST  list of excluded directories
		-np, --no-parent                 don't ascend to the parent directory
	
		Email bug reports, questions, discussions to <bug-wget@gnu.org>
		and/or open issues at https://savannah.gnu.org/bugs/?func=additem&group=wget.`
}
