package sshoneypot

import (
	"errors"
	"fmt"
	"github.com/bytedream/sshoneypot/sshoneypot/handler"
	"github.com/bytedream/sshoneypot/sshoneypot/info"
	"golang.org/x/crypto/ssh"
	"golang.org/x/crypto/ssh/terminal"
	"io"
	"net"
	"os"
	"path/filepath"
	"runtime"
	"sort"
	"strings"
)

const keyTab = 9

type SSHoneypot struct {
	AllowedUsers []string
	AllowedPasswords []string

	Panic   bool
	Logger  *Logging
	Handler map[string]info.Handler
	FS      *info.FileSystem

	Port uint16
	ServerConfig *ssh.ServerConfig
}

func (honeypot SSHoneypot) AddHandler(name string, handler info.Handler, aliases... string) error {
	for _, name := range append([]string{name}, aliases...) {
		if _, exist := honeypot.Handler[name]; exist {
			return errors.New("The handler name '" + name + "' already exists")
		}
		honeypot.Handler[name] = handler
	}
	return nil
}

type HandlerFunc func(info *info.Info)

//simpleHandler is a easy implementation of Handler / SimpleHandler for the AddHandlerFunc method below
type simpleHandler struct {
	info.Handler

	HandlerName string
	handleFunc HandlerFunc
}

func (sh simpleHandler) AutoCompleteOptions() []string {
	return make([]string, 0)
}
func (sh simpleHandler) AutoCompleteRequest(info *info.Info) (accepted bool) {
	return false
}
func (sh simpleHandler) Handle(info *info.Info) {
	sh.handleFunc(info)
}
func (sh simpleHandler) Kill() {}

func (honeypot SSHoneypot) AddHandlerFunc(name string, handler HandlerFunc, aliases... string) error {
	return honeypot.AddHandler(name, simpleHandler{handleFunc: handler}, aliases...)
}

// Serve starts the ssh server and serve it until the program ends
func (honeypot SSHoneypot) Serve() error {
	// starts the ssh listener
	listener, err := net.Listen("tcp", fmt.Sprintf(":%d", honeypot.Port))
	if err != nil {
		return err
	}
	defer listener.Close()

	for {
		conn, err := listener.Accept()
		if err != nil {
			honeypot.Logger.Warnf("Failed to accept new client %s: %s\n", conn.RemoteAddr(), err)
			continue
		}
		go honeypot.handleConn(conn)
	}
}

func (honeypot SSHoneypot) handleConn(conn net.Conn) {
	defer conn.Close()

	serverConn, channels, requests, err := ssh.NewServerConn(conn, honeypot.ServerConfig)
	if err != nil {
		honeypot.Logger.Warnf("Failed to create new connection with %s: %s\n", conn.RemoteAddr(), err)
		return
	}
	term := &info.Termina{}
	go handleRequest(requests, term)
	for channel := range channels {
		go honeypot.handleChannel(conn, channel, serverConn, term)
	}
	honeypot.Logger.Infof("%s closed the connection\n", conn.RemoteAddr())
}

func (honeypot SSHoneypot) handleChannel(conn net.Conn, newChannel ssh.NewChannel, serverConn *ssh.ServerConn, term *info.Termina) {
	channel, request, err := newChannel.Accept()
	if err != nil {
		return
	}

	for {
		info.NewTerminal(channel).ReadLine()
	}

	term.Terminal = terminal.NewTerminal(channel, "")
	term.SetPrompt(fmt.Sprintf("%s:%s $ ", serverConn.User(), "/"))
	term.AutoCompleteCallback = func(line string, pos int, key rune) (newLine string, newPos int, ok bool) {
		if key == keyTab {
			if command, err := parseCommand(line); len(command) > 0 && err == nil {
				if cmd, exists := honeypot.Handler[command[0]]; exists {
					cmdInfo := &info.Info{Conn: conn, Terminal: term, FS: honeypot.FS, User: serverConn, FullArgs: command, Command: command[0], Args: command[1:], Stdout: true, Stderr: true}
					if cmd.AutoCompleteRequest != nil && !cmd.AutoCompleteRequest(cmdInfo){
						if options := cmd.AutoCompleteOptions(); options != nil && len(options) == 0 {
							newLine, newPos, ok = honeypot.fileAutoComplete(cmdInfo, line)
						} else if len(options) == 1 {
							newLine = line + options[0]
						}
					}
				}
			}
		}
		return
	}
	go handleRequest(request, term)
	for {
		line, err := term.ReadLine()
		if err != nil {
			if err == io.EOF {
				if honeypot.OnClose(channel) {
					return
				}
			}
		} else {
			fmt.Println(line)
			/*if !honeypot.handleLine(line, conn, channel, serverConn, term) {
				if honeypot.OnClose(channel) {
					return
				}
			}*/
		}
	}
}

func (honeypot SSHoneypot) handleLine(line string, conn net.Conn, channel ssh.Channel, serverConn *ssh.ServerConn, term *info.Termina) string {
	command, err := parseCommand(line)
	if err == nil && len(command) > 0 {
		clientInfo := &info.Info{Conn: conn, Terminal: term, FS: honeypot.FS, User: serverConn, FullArgs: command, Command: command[0], Args: command[1:], Stdout: true, Stderr: true}
		var redirect bool
		for i, arg := range clientInfo.Args {
			if redirect {
				redirect = false
				clientInfo.Args = append(clientInfo.Args[:i], clientInfo.Args[i+1:]...)
			}
			switch arg {
			case "2>", "&>":
				clientInfo.Stderr = false
				fallthrough
			case ">", "1>", ">>":
				clientInfo.Stdout = false

				clientInfo.Args = append(clientInfo.Args[:i], clientInfo.Args[i+1:]...)
			}
		}
		if redirect {
			clientInfo.Errorln("-bash: syntax error near unexpected token 'newline'")
			return ""
		}

		finished := make(chan bool, 1)
		newLine := ""

		go func() {
			for {
				var buffer = make([]byte, 512)
				n, _ := channel.Read(buffer)
				select {
				case <- finished:
					newLine = string(buffer[:n])
					return
				default:
					if buffer[0] == 3 {
						clientInfo.Stderr = false
						clientInfo.Stdout = false
						finished <- true
					}
				}
			}
		}()

		if commandHandler, ok := honeypot.Handler[command[0]]; ok {
			if honeypot.Panic {
				commandHandler.Handle(clientInfo)
			} else {
				func() {
					defer func() {
						if r := recover(); r != nil {
							buffer := make([]byte, 2048)
							written := runtime.Stack(buffer, false)
							honeypot.Logger.Warnf("Unexpected panic: %s\n%s", r, buffer[:written])
						}
					}()
					commandHandler.Handle(clientInfo)
				}()
			}

			// the following part is the implementation of ctrl + c / ctrl + d when a process is running.
			// sadly the ssh library doesn't support ctrl + c / some other keys to handle their input directly,
			// so a second empty terminal has to be opened.
			// and because of this ugly implementation, the key up / down buttons to get the recent entered command are not working :/



			return newLine
		} else {
			handler.Default(clientInfo)
		}
	}
	return ""
}

func (honeypot SSHoneypot) fileAutoComplete(cmdInfo *info.Info, line string) (newLine string, newPos int, ok bool) {
	var names []string

	if len(cmdInfo.Args) > 0 {
		var dir info.Directory
		search := cmdInfo.Args[len(cmdInfo.Args) - 1]

		if strings.HasPrefix(search, string(os.PathSeparator)) {
			dir, _ = honeypot.FS.GetExplicitDirectory("/", false)
		} else {
			dir = honeypot.FS.Cwd()
		}

		for _, file := range dir.Files {
			if strings.HasPrefix(file.Name, search) && file.Name != search {
				names = append(names, file.Name)
			}
		}
		if len(names) == 1 && strings.HasPrefix(search, string(os.PathSeparator)) {
			names[0] = filepath.Join(string(os.PathSeparator), names[0])
		} else if len(names) == 0 && strings.Contains(search, string(os.PathSeparator)) {
			parent, _ := filepath.Split(search)
			if dir, ok := honeypot.FS.GetExplicitDirectory(parent, false); ok {
				for _, file := range dir.Files {
					if strings.HasPrefix(filepath.Join(parent, file.Name), search) && file.Name != search {
						names = append(names, file.Name)
					}
				}
			}
			if len(names) == 1 {
				names[0] = filepath.Join(parent, names[0]) + string(os.PathSeparator)
			}
		}
	} else {

		dir := honeypot.FS.Cwd()
		for _, file := range dir.Files {
			names = append(names, file.Name)
		}
	}

	if len(names) == 0 {
		return
	} else if len(names) == 1 {
		ok = true
		if len(cmdInfo.Args) == 0 {
			newLine += names[0]
		} else {
			newLine = line[:strings.LastIndex(line, " ")] + " " + names[0]
		}
		newPos = len(newLine)
	} else {
		honeypot.OptionsAutoComplete(names, cmdInfo, line)
	}
	return
}

// OptionsAutoComplete will show all given options in the console.
// e.g. all files in the current directory when double pressing `tab` after the `ls` command
func (honeypot SSHoneypot) OptionsAutoComplete(options []string, info *info.Info, line string) {
	sort.Strings(options)

	var length int
	for _, option := range options {
		if l := len(option); length < l {
			length = l
		}
	}
	length++

	var builder strings.Builder
	var perline int
	if info.Terminal.Width > 0 {
		perline = int(info.Terminal.Width) / length
	} else {
		perline = 80 / length
	}
	builder.WriteString(info.Terminal.Prompt + line + "\n")
	for i, option := range options {
		if i % perline == 0 && i != 0 {
			builder.WriteString("\n")
		}
		builder.WriteString(option + strings.Repeat(" ", length - len(option)))
	}
	builder.WriteString("\n")

	info.Write(builder.String())
}

// --- The following methods were created to allow user to manipulate the ssh process as they wish --- //

// OnClose gets called when the client send an exit signal.
// If the return bool is true, the channel will be closed an the client disconnected
func (honeypot SSHoneypot) OnClose(channel ssh.Channel) (close bool) {
	channel.SendRequest("exit-status", false, ssh.Marshal(struct {Status uint32}{Status: 0}))
	channel.Close()
	return true
}

// --- //

func DefaultSSHoneypot(fs *info.FileSystem, key ssh.Signer) SSHoneypot {
	serverConfig := &ssh.ServerConfig{
		PasswordCallback: func(conn ssh.ConnMetadata, password []byte) (*ssh.Permissions, error) {
			return nil, nil
		},
		ServerVersion: "SSH-2.0-OpenSSH_8.5",
	}
	serverConfig.AddHostKey(key)

	honeypot := SSHoneypot{
		Logger: &Logging{enable: true},
		Port: 2222,
		ServerConfig: serverConfig,

		Handler: map[string]info.Handler{},

		FS: fs,
	}
	honeypot.AddHandler("cat", handler.Cat{}, "/bin/cat")
	honeypot.AddHandler("cd", handler.CD{})
	honeypot.AddHandler("ls", handler.LS{}, "/bin/ls")
	honeypot.AddHandler("mkdir", handler.Mkdir{}, "/bin/mkdir")
	honeypot.AddHandler("rm", handler.RM{}, "/bin/rm")

	honeypot.AddHandlerFunc("clear", handler.Clear, "/bin/clear")
	honeypot.AddHandlerFunc("echo", handler.Echo, "/bin/echo")

	honeypot.AddHandlerFunc("ifconfig", handler.Ifconfig, "/sbin/ifconfig")
	honeypot.AddHandler("ping", handler.Ping{}, "/bin/ping", "/bin/ping4", "/bin/ping6")

	return honeypot
}
